# pylint: disable=import-error
import json

import fischertechnik
import fischertechnik.factories as txt_factory

# pylint: enable=import-error
import txtapi
from txtapi.models.controller import Controller


class Txt:

    MAX_RETRY = 50
    __instance = None
    _controllers = [None] * 10
    _camera = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if Txt.__instance == None:
            Txt()
        return Txt.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if Txt.__instance != None:
            raise Exception("This class is a singleton!")
        else:
            Txt.__instance = self
            txt_factory.init_controller_factory()
            txt_factory.init_input_factory()
            txt_factory.init_output_factory()
            txt_factory.init_motor_factory()
            txt_factory.init_counter_factory()
            txt_factory.init_usb_factory()
            txt_factory.init_camera_factory()
            txt_factory.init_servomotor_factory()


    def init_controller(self, controller_id):
        if controller_id < len(self._controllers) and self._controllers[controller_id] is None:
            self._controllers[controller_id] = txt_factory.controller_factory.create_graphical_controller(
                controller_id)


    def create_counter(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            return txt_factory.counter_factory.create_counter(self._controllers[controller_id], idx)
        return None


    def create_input(self, controller_id, idx, input_device):
        if self._controllers[controller_id] is not None:
            if input_device.device == "MINI_SWITCH":
                return txt_factory.input_factory.create_mini_switch(self._controllers[controller_id], idx)
            elif input_device.device == "PHOTO_RESISTOR":
                return txt_factory.input_factory.create_photo_resistor(self._controllers[controller_id], idx)
            elif input_device.device == "NTC_RESISTOR":
                return txt_factory.input_factory.create_ntc_resistor(self._controllers[controller_id], idx)
            elif input_device.device == "ULTRASONIC_DISTANCE_METER":
                return txt_factory.input_factory.create_ultrasonic_distance_meter(self._controllers[controller_id], idx)
            elif input_device.device == "PHOTO_TRANSISTOR":
                return txt_factory.input_factory.create_photo_transistor(self._controllers[controller_id], idx)
            elif input_device.device == "COLOR_SENSOR":
                return txt_factory.input_factory.create_color_sensor(self._controllers[controller_id], idx)
            elif input_device.device == "TRAIL_FOLLOWER":
                return txt_factory.input_factory.create_trail_follower(self._controllers[controller_id], idx)
            else:
                return txt_factory.input_factory.create_mini_switch(self._controllers[controller_id], idx)
        return None


    def get_motor(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            if self._controllers[controller_id].get_motor(idx) == None:
                self._controllers[controller_id].set_motor(idx, txt_factory.motor_factory.create_motor(
                    self._controllers[controller_id], idx))
            return self._controllers[controller_id].get_motor(idx)
        return None


    def get_servomotor(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            if self._controllers[controller_id].get_servomotor(idx) == None:
                self._controllers[controller_id].set_servomotor(idx, txt_factory.servomotor_factory.create_servomotor(
                    self._controllers[controller_id], idx))
            return self._controllers[controller_id].get_servomotor(idx)
        return None


    def get_output(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            if self._controllers[controller_id].get_output(idx) == None:
                self._controllers[controller_id].set_output(idx, txt_factory.output_factory.create_lamp(
                    self._controllers[controller_id], idx))
            return self._controllers[controller_id].get_output(idx)
        return None


    def get_input(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            return self._controllers[controller_id].get_input(idx)
        return None


    def get_inputs(self, controller_id):
        if self._controllers[controller_id] is not None:
            return self._controllers[controller_id]._input
        return None


    def get_counter(self, controller_id, idx):
        if self._controllers[controller_id] is not None:
            return self._controllers[controller_id].get_counter(idx)
        return None


    def get_counters(self, controller_id):
        if self._controllers[controller_id] is not None:
            return self._controllers[controller_id]._counter
        return None


    def get_camera(self):
        if self._camera is None:
            self._camera = txt_factory.usb_factory.create_camera(self.master_controller(), 1)
        return self._camera


    def add_camera_image_recognition_config(self, config):

        camera = self.get_camera()
        camera.clear_blocked_areas()
        camera.clear_detectors()

        if config.blocked_areas is not None:
            for area in config.blocked_areas:
                camera.add_blocked_area(area)

        if config.motion_detectors is not None:
            for motion_config in config.motion_detectors:
                motion_detector = txt_factory.camera_factory.create_motion_detector(
                    motion_config.area.x,
                    motion_config.area.y,
                    motion_config.area.width,
                    motion_config.area.height,
                    motion_config.tolerance
                )
                motion_detector.set_identifier(motion_config.name)
                camera.add_detector(motion_detector)

        if config.color_detectors is not None:
            for color_config in config.color_detectors:
                color_detector = txt_factory.camera_factory.create_color_detector(
                    color_config.area.x,
                    color_config.area.y,
                    color_config.area.width,
                    color_config.area.height,
                    color_config.contrast
                )
                color_detector.set_identifier(color_config.name)
                camera.add_detector(color_detector)

        if config.ball_detectors is not None:
            for ball_config in config.ball_detectors:
                ball_detector = txt_factory.camera_factory.create_ball_detector(
                    ball_config.area.x,
                    ball_config.area.y,
                    ball_config.area.width,
                    ball_config.area.height,
                    ball_config.min_ball_diameter,
                    ball_config.max_ball_diameter,
                    ball_config.start_range_value,
                    ball_config.end_range_value,
                    ball_config.rgb,
                    ball_config.tolerance
                )
                ball_detector.set_identifier(ball_config.name)
                camera.add_detector(ball_detector)

        if config.line_detectors is not None:
            for line_config in config.line_detectors:
                line_detector = txt_factory.camera_factory.create_line_detector(
                    line_config.area.x,
                    line_config.area.y,
                    line_config.area.width,
                    line_config.area.height,
                    line_config.min_line_width,
                    line_config.max_line_width,
                    line_config.start_range_value,
                    line_config.end_range_value,
                    line_config.number_of_lines,
                    line_config.invert
                )
                line_detector.set_identifier(line_config.name)
                camera.add_detector(line_detector)


    def start_camera(self, config):
        self.stop_camera()
        camera = self.get_camera()
        camera.set_fps(config.fps)
        camera.set_width(config.width)
        camera.set_height(config.height)
        camera.set_debug(config.debug)
        camera.set_rotate(config.rotate)
        camera.clear_blocked_areas()
        camera.clear_detectors()
        camera.start()


    def stop_camera(self):
        if self._camera is not None:
            self._camera.stop()


    def discover(self):
        controllers = []
        if (txt_factory.api_model == 1):
            name = self.master_controller()._txt.getDevicename()
            firmware = self.master_controller()._txt.getFirmwareVersion()
            version = self.master_controller()._txt.getVersionNumber()
            txtapi_version = txtapi.__version__
            controller_lib_version = fischertechnik.__version__
            controllers.append(Controller(name=name, firmware=firmware, version=version, api_version=txtapi_version,
                                          controller_lib_version=controller_lib_version))
        else:
            controllerInfoList = self.master_controller()._txt.getControllerInfoList()
            for i in range(len(controllerInfoList)):
                if i == 0:
                    controllers.append(Controller(name=controllerInfoList[i].name,
                                                  serial_number=controllerInfoList[i].sn,
                                                  firmware=self.master_controller()._txt.firmware,
                                                  version=self.master_controller()._txt.version,
                                                  api_version=txtapi.__version__,
                                                  controller_lib_version=fischertechnik.__version__))
                else:
                    controller = self.master_controller()._txt.getController(controllerInfoList[i])
                    controllers.append(Controller(name=controllerInfoList[i].name,
                                                  serial_number=controllerInfoList[i].sn,
                                                  firmware=controller.firmware,
                                                  version=controller.version,
                                                  api_version="None",
                                                  controller_lib_version="None"
                                                  ))
        return controllers


    def master_controller(self):
        if self._controllers[0] is None:
            self.init_controller(0)
        return self._controllers[0]


    def stop(self):
        self.stop_camera()
        for idx in range(len(self._controllers)):
            if self._controllers[idx] is not None:
                self._controllers[idx].cleanup()
