import os
import socket
from subprocess import Popen
from typing import List
from time import sleep
import txtapi.handlers.workspace_handler as workspace_handler
from txtapi.utility.debugger.rpdb_helper import EOT
from txtapi.utility.debugger.debugger_commands import Command


class FtDebugger():
    min_port = 30100
    max_ports = 100
    port_counter = 0
    max_connection_attempts = 5
    buflen = 1024
    sock_timeout = 20

    def __init__(self, workspace_name: str) -> None:
        self.workspace_name = workspace_name
        self.increment_port_num()
        self.is_socket_open = False
        self.rpdb_executable_path = workspace_handler.get_file_path(workspace_name, workspace_name + ".py")
        if not os.path.exists(self.rpdb_executable_path):
            raise FileNotFoundError("Failed to start process: " + self.rpdb_executable_path + " does not exist.")


    def increment_port_num(self):
        self.port = FtDebugger.min_port + (FtDebugger.port_counter % FtDebugger.max_ports)
        FtDebugger.port_counter += 1


    def open_subprocess(self, args):
        popen_args = [
            "python3",
            "-m",
            "txtapi.utility.debugger.rpdb_helper",
            str(self.port),
            self.rpdb_executable_path,
        ]
        if args is not None:
            popen_args.extend(args)
        path = workspace_handler.get_file_path(self.workspace_name, "")
        # specify PYTHONPATH so that imports still work
        my_env = os.environ.copy()
        try:
            python_path = my_env.get("PYTHONPATH") + ':' + path
        except:
            python_path = path
        my_env.update({"PYTHONPATH": python_path})
        self.process = Popen(popen_args, env=my_env)


    def start(self, args) -> str:
        self.open_subprocess(args)
        for x in range(FtDebugger.max_connection_attempts):
            # wait a second so the process has time to start
            sleep(1)
            if self.process.poll() is not None:
                # if the process is not running, it probably returned with EADDRINUSE.
                self.increment_port_num()
                self.open_subprocess(args)
            else:
                try:
                    self.sock = socket.create_connection(("localhost", self.port))
                    self.sock.settimeout(FtDebugger.sock_timeout)
                    print("FtDebugger: debugging {0} on port {1}.".format(self.workspace_name, self.port))
                    self.is_socket_open = True
                    return self._recv_rpdb_data()
                except ConnectionRefusedError:
                    pass
        self.process.terminate()
        return "Could not start debugger after {} tries.".format(FtDebugger.max_connection_attempts)


    def _recv_rpdb_data(self) -> str:
        try:
            debugger_response = ""
            while True:
                data = self.sock.recv(FtDebugger.buflen).decode("utf-8")
                debugger_response += data
                if not data:
                    self._close_socket()
                    return debugger_response
                if debugger_response.endswith(EOT):
                    return debugger_response
        except Exception as e:
            self._close_socket()
            raise Exception(e)


    def _close_socket(self):
        self.is_socket_open = False
        self.sock.close()


    def send_command(self, command: str) -> str:
        self.sock.sendall(bytes(command + "\n", "utf-8"))
        return self._recv_rpdb_data()


    def send_clear_command(self) -> str:
        self.sock.sendall(bytes(Command.CLEAR.value + "\n", "utf-8"))
        debugger_response = ""
        while True:
            data = self.sock.recv(FtDebugger.buflen).decode("utf-8")
            debugger_response += data
            if not data:
                self._close_socket()
                return debugger_response
            if "Clear all breaks?" in debugger_response:
                return self.send_command(Command.YES.value)
