#!/usr/bin/env python
import fcntl
import os
import queue
from threading import Thread


class WatchProcess:
    stop_output = False
    _thread = None
    _running = True
    _message_queue = queue.Queue()

    # Here will be the instance stored.
    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if WatchProcess.__instance == None:
            WatchProcess()
        return WatchProcess.__instance

    def __init__(self):
        """ Virtually private constructor. """
        if WatchProcess.__instance != None:
            raise Exception("This class is a Singleton!")
        else:
            WatchProcess.__instance = self

    def stop(self):
        if self._thread is not None:
            self._running = False
            if self._thread.is_alive() is True:
                self._thread.join()
                self.stop_output = True
        self._thread = None

    def start(self, process):
        self.stop()
        self.stop_output = False
        self._running = True
        self._thread = Thread(target=self._watch_output, args=[process])
        self._thread.daemon = True
        self._thread.start()

    def clear(self):
        self._message_queue = queue.Queue()

    def send_started(self):
        if self._message_queue is not None:
            self._message_queue.put({'type': 'started', 'data': ''}, True)

    def send_finished(self):
        if self._message_queue is not None:
            self._message_queue.put({'type': 'finished', 'data': ''}, True)

    def send_clear(self):
        self._message_queue.put({'type': 'clear', 'data': ''}, True)

    def send_data(self, data):
        if self._message_queue is not None:
            if data is not None:
                data_array = data.splitlines()
                if len(data_array) > 0:
                    self._message_queue.put({'type': 'text', 'data': data_array}, True)

    def has_messages(self):
        return not self._message_queue.empty()

    def get_message(self, block=False, timeout=None):
        return self._message_queue.get(block, timeout)

    def task_done(self):
        self._message_queue.task_done()

    def join(self):
        self._message_queue.join()

    def _watch_output(self, process):
        self.clear()
        self.send_clear()
        self.send_started()
        while self._running:
            output = self._non_block_read(process.stdout)
            if output:
                data = str(output.decode("utf-8"))
                self.send_data(data)
            else:
                if process.poll() is not None:
                    # process is terminated
                    break
        self.send_finished()
        self.stop_output = True

    def _non_block_read(self, output):
        # make stdin a non-blocking file
        fd = output.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        try:
            return output.readline()
        except:
            return ""
